@@ -6,12 +6,13 @@ import pandas as pd
 import numpy as np
 import inspect
 import re
+import warnings
 
 
 class AgGridReturn(Mapping):
     """
     Container for AgGrid component response data.
-    
+
     Provides easy access to grid data, selected rows, grid state, and other
     information returned by the AgGrid component.
     """
@@ -19,28 +20,28 @@ class AgGridReturn(Mapping):
     def __init__(
         self,
         originalData,
-        gridOptions=None,
+        grid_response={},
         data_return_mode=DataReturnMode.AS_INPUT,
-        try_to_convert_back_to_original_types=True,
         conversion_errors="coerce",
+        frame_dtypes=None,
     ) -> None:
         super().__init__()
 
         # Configuration
         self._original_data = originalData
-        self._try_convert_types = try_to_convert_back_to_original_types
-        self._conversion_errors = conversion_errors
         self._data_return_mode = data_return_mode
-        
+        self._conversion_errors = conversion_errors
+
         # State
-        self._component_value_set = False
-        self.__dict__["grid_response"] = {} #{"gridOptions": gridOptions or {}}
+        self._component_value_set = grid_response is True
+        self.__dict__["grid_response"] = grid_response
+        self.frame_dtypes = frame_dtypes
 
     def _set_component_value(self, component_value):
         """Set the response value from the AgGrid component."""
         self._component_value_set = True
         self.__dict__["grid_response"] = component_value
-        
+
         # Ensure gridOptions is a dict
         grid_options = self.__dict__["grid_response"].get("gridOptions")
         if grid_options and not isinstance(grid_options, dict):
@@ -49,7 +50,7 @@ class AgGridReturn(Mapping):
     # ==========================================
     # Basic Properties - Direct Grid Response Access
     # ==========================================
-    
+
     @property
     def grid_response(self):
         """Raw response from component."""
@@ -90,76 +91,116 @@ class AgGridReturn(Mapping):
     # ==========================================
 
     def _convert_column_types(self, data):
-        """Convert DataFrame columns back to their original types."""
-        if not self._try_convert_types:
-            return data
-            
-        original_types = self.grid_response.get("originalDtypes", {})
-        
-        # Handle both dict and list formats for originalDtypes
-        if isinstance(original_types, dict):
-            original_types.pop("::auto_unique_id::", None)  # Remove if exists
-        elif isinstance(original_types, list):
-            # Convert list to dict format if needed (legacy compatibility)
-            if data.columns.tolist():
-                original_types = dict(zip(data.columns.tolist(), original_types))
-                original_types.pop("::auto_unique_id::", None)
-        else:
-            original_types = {}
-        
-        # Group columns by type for batch processing
-        type_groups = {
-            'numeric': [k for k, v in original_types.items() if v in ["i", "u", "f"]],
-            'text': [k for k, v in original_types.items() if v in ["O", "S", "U"]],
-            'date': [k for k, v in original_types.items() if v == "M"],
-            'timedelta': [k for k, v in original_types.items() if v == "m"]
-        }
-        
-        # Convert numeric columns
-        if type_groups['numeric']:
-            data.loc[:, type_groups['numeric']] = data.loc[:, type_groups['numeric']].apply(
-                pd.to_numeric, errors=self._conversion_errors
+        """Convert DataFrame columns back to their original types.
+
+        Args:
+            data: DataFrame with columns to convert
+
+        Returns:
+            DataFrame with columns converted to original types
+        """
+        converted_columns = []
+
+        for col_name in data.columns:
+            column = data[col_name]
+
+            # Keep UI-created columns as-is (they don't have original dtypes)
+            if col_name not in self.frame_dtypes:
+                converted_columns.append(column)
+                continue
+
+            original_dtype = self.frame_dtypes[col_name]
+            dtype_kind = original_dtype.kind
+
+            # Convert based on original dtype kind
+            if dtype_kind == "i":  # Integer
+                converted_columns.append(
+                    self._convert_to_integer(column)
+                )
+            elif dtype_kind == "f":  # Float
+                converted_columns.append(
+                    pd.to_numeric(column, errors=self._conversion_errors).astype(
+                        original_dtype, copy=False
+                    )
+                )
+            elif dtype_kind in ("O", "S", "U"):  # Object/String/Unicode
+                converted_columns.append(
+                    column.astype(original_dtype, copy=False)
+                )
+            elif dtype_kind == "M":  # Datetime
+                converted_columns.append(
+                    pd.to_datetime(column, errors=self._conversion_errors).astype(
+                        original_dtype, copy=False
+                    )
+                )
+            elif dtype_kind == "m":  # Timedelta
+                converted_columns.append(
+                    self._convert_to_timedelta(column, original_dtype)
+                )
+            else:  # Other types
+                converted_columns.append(
+                    column.astype(original_dtype)
+                )
+
+        return pd.concat(converted_columns, axis=1, copy=False)
+
+    def _convert_to_integer(self, column):
+        """Convert column to Int64, falling back to Float64 on errors.
+
+        Args:
+            column: Series to convert
+
+        Returns:
+            Series converted to Int64 or Float64
+        """
+        try:
+            return pd.to_numeric(
+                column, downcast="integer", errors="coerce"
+            ).astype("Int64", copy=False)
+        except TypeError:
+            warnings.warn(
+                f"Error casting {column.name} to Int64. Falling back to Float64"
             )
-        
-        # Convert text columns
-        if type_groups['text']:
-            data.loc[:, type_groups['text']] = data.loc[:, type_groups['text']].map(
-                lambda x: np.nan if x is None else str(x)
+            return pd.to_numeric(column, errors="coerce").astype(
+                "Float64", copy=False
             )
-        
-        # Convert date columns
-        if type_groups['date']:
-            data.loc[:, type_groups['date']] = data.loc[:, type_groups['date']].apply(
-                pd.to_datetime, errors=self._conversion_errors
-            )
-        
-        # Convert timedelta columns
-        if type_groups['timedelta']:
-            def safe_timedelta(s):
-                try:
-                    return pd.Timedelta(s)
-                except (ValueError, TypeError, pd.errors.OutOfBoundsDatetime):
-                    return s
-            
-            data.loc[:, type_groups['timedelta']] = data.loc[:, type_groups['timedelta']].apply(
-                safe_timedelta
-            )
-        
-        return data
+
+    def _convert_to_timedelta(self, column, original_dtype):
+        """Convert column to timedelta, handling errors gracefully.
+
+        Args:
+            column: Series to convert
+            original_dtype: Target dtype for the column
+
+        Returns:
+            Series converted to timedelta
+        """
+        def safe_timedelta(value):
+            """Convert value to Timedelta, returning original on error."""
+            try:
+                return pd.Timedelta(value)
+            except (ValueError, TypeError, pd.errors.OutOfBoundsDatetime):
+                return value
+
+        return column.apply(safe_timedelta).astype(original_dtype, copy=False)
 
     def _create_dataframe_from_nodes(self, nodes):
         """Create a DataFrame from grid nodes."""
         # Extract data from non-group nodes
-        data = pd.DataFrame([
-            n.get("data", {}) for n in nodes 
-            if not n.get("group", False)
-        ])
-        
+        data = pd.DataFrame(
+            [n.get("data", {}) for n in nodes if not n.get("group", False)],
+            dtype=object,
+        )
+
         # Set index from auto_unique_id if available
         if "::auto_unique_id::" in data.columns:
             data.index = pd.Index(data["::auto_unique_id::"], name="index")
-        
-        return self._convert_column_types(data)
+            # Remove the internal column - it's only used for indexing
+            data = data.drop(columns=["::auto_unique_id::"])
+
+        if self.frame_dtypes is not None:
+            data = self._convert_column_types(data)
+        return data
 
     def _process_grouped_response(self, nodes):
         """Process nodes with grouping information."""
@@ -170,7 +211,7 @@ class AgGridReturn(Mapping):
                 parent_path = node.get("parentPath", "")
                 row_data = {**node.get("data", {}), "parentPath": parent_path}
                 data_rows.append(row_data)
-        
+
         # Set index and clean up
         data = pd.DataFrame(data_rows)
         if "::auto_unique_id::" in data.columns:
@@ -178,7 +219,7 @@ class AgGridReturn(Mapping):
             # Apply filtering and sorting if needed
             data = self._apply_filtering_and_sorting(data, only_selected=False)
             data.index.name = ""
-        
+
         # Group by parent path and parse AG-Grid IDs for meaningful group names
         # Use sort=False to preserve original order and improve performance
         groups = []
@@ -186,79 +227,79 @@ class AgGridReturn(Mapping):
             group_key = self._parse_aggrid_group_ids(parent_path)
             clean_data = group_data.drop("parentPath", axis=1)
             groups.append({group_key: clean_data})
-        
+
         return groups
 
     def _parse_aggrid_group_ids(self, parent_path: str) -> tuple:
         """Parse AG-Grid auto-generated IDs to extract meaningful group names.
-        
+
         Based on actual observed structure:
         Example: "ROOT_NODE_ID.row-group-sport-Swimming.row-group-sport-Swimming-athlete-Michael Phelps"
-        
+
         Pattern analysis:
         - Level 1: "row-group-sport-Swimming" -> key is "Swimming"
         - Level 2: "row-group-sport-Swimming-athlete-Michael Phelps" -> new key is "Michael Phelps"
-        
+
         Each level adds: -{colId}-{key} to the previous path
         We need to extract just the keys in order: ("Swimming", "Michael Phelps")
         """
         if not parent_path:
             return ()
-        
+
         # Remove ROOT_NODE_ID prefix if present
         if parent_path.startswith("ROOT_NODE_ID."):
             parent_path = parent_path[13:]  # len("ROOT_NODE_ID.") = 13
-        
+
         # Split by dots to get each level
-        parts = parent_path.split('.')
+        parts = parent_path.split(".")
         group_keys = []
-        
+
         for i, part in enumerate(parts):
-            if part.startswith('row-group-'):
+            if part.startswith("row-group-"):
                 # Remove 'row-group-' prefix
                 content = part[10:]  # len('row-group-') = 10
-                
+
                 if not content:
                     continue
-                
+
                 if i == 0:
                     # First level: row-group-{colId}-{key}
                     # Find the first dash and take everything after it
-                    first_dash = content.find('-')
+                    first_dash = content.find("-")
                     if first_dash > 0:
-                        key = content[first_dash + 1:]
+                        key = content[first_dash + 1 :]
                         group_keys.append(key)
                 else:
                     # Subsequent levels contain the full path: {previousPath}-{colId}-{key}
                     # We need to find what's new compared to the previous level
-                    
+
                     # Get the previous part to compare
-                    prev_part = parts[i-1]
-                    if prev_part.startswith('row-group-'):
+                    prev_part = parts[i - 1]
+                    if prev_part.startswith("row-group-"):
                         prev_content = prev_part[10:]
-                        
+
                         # The current content should start with prev_content
                         # followed by -{colId}-{key}
                         if content.startswith(prev_content):
                             # Extract the new part: -{colId}-{key}
-                            new_part = content[len(prev_content):]
-                            if new_part.startswith('-'):
+                            new_part = content[len(prev_content) :]
+                            if new_part.startswith("-"):
                                 new_part = new_part[1:]  # Remove leading dash
-                                
+
                                 # Find the next dash (after colId) and extract key
-                                dash_pos = new_part.find('-')
+                                dash_pos = new_part.find("-")
                                 if dash_pos > 0:
-                                    key = new_part[dash_pos + 1:]
+                                    key = new_part[dash_pos + 1 :]
                                     group_keys.append(key)
                                 else:
                                     # No dash found, the whole thing is the key
                                     group_keys.append(new_part)
                         else:
                             # Fallback: extract the last key-like segment
-                            segments = content.split('-')
+                            segments = content.split("-")
                             if len(segments) >= 2:
                                 group_keys.append(segments[-1])
-        
+
         return tuple(group_keys)
 
     def _get_data(self, only_selected=False):
@@ -267,7 +308,7 @@ class AgGridReturn(Mapping):
             return None if only_selected else self._original_data
 
         nodes = self.grid_response.get("nodes", [])
-        
+
         # Filter to selected nodes if requested
         if only_selected:
             nodes = [n for n in nodes if n.get("isSelected", False)]
@@ -275,14 +316,17 @@ class AgGridReturn(Mapping):
                 return None
 
         # Handle DataFrame data
-        if isinstance(self._original_data, pd.DataFrame) and not self._original_data.empty:
+        if (
+            isinstance(self._original_data, pd.DataFrame)
+            and not self._original_data.empty
+        ):
             data = self._create_dataframe_from_nodes(nodes)
             return self._apply_filtering_and_sorting(data, only_selected)
-        
+
         # Handle JSON/string data or empty DataFrame
         if self._should_return_json_data():
             return self._create_json_response(nodes)
-        
+
         return self._original_data if not only_selected else None
 
     def _apply_filtering_and_sorting(self, data, only_selected):
@@ -299,22 +343,22 @@ class AgGridReturn(Mapping):
             reindex_ids = pd.Index(reindex_ids)
             if only_selected:
                 reindex_ids = reindex_ids.intersection(data.index)
-            
+
             data = data.reindex(index=reindex_ids).reset_index(drop=True)
-            
+
             # Remove auto_unique_id column if present
             columns = [col for col in data.columns if col != "::auto_unique_id::"]
             return data[columns]
-        
+
         return data
 
     def _should_return_json_data(self):
         """Check if we should return JSON data instead of DataFrame."""
         data = self._original_data
         return (
-            (isinstance(data, str) and self._is_valid_json(data)) or
-            (isinstance(data, pd.DataFrame) and data.empty) or
-            (data is None)
+            (isinstance(data, str) and self._is_valid_json(data))
+            or (isinstance(data, pd.DataFrame) and data.empty)
+            or (data is None)
         )
 
     def _is_valid_json(self, data_str):
@@ -335,12 +379,18 @@ class AgGridReturn(Mapping):
             filter_ids = None
 
         sorted_nodes = sorted(nodes, key=lambda n: n.get("rowIndex", 0))
-        
+
         if filter_ids:
             data_list = [n["data"] for n in sorted_nodes if n["id"] in filter_ids]
         else:
             data_list = [n["data"] for n in sorted_nodes]
-        
+
+        # Remove internal columns from each data item
+        data_list = [
+            {k: v for k, v in item.items() if not k.startswith("::")}
+            for item in data_list
+        ]
+
         return json.dumps(data_list)
 
     # ==========================================
@@ -363,7 +413,7 @@ class AgGridReturn(Mapping):
             return [{(): pd.DataFrame()}]
 
         nodes = self.grid_response.get("nodes", [])
-        
+
         if only_selected:
             # Default is True because AgGrid sets undefined for half-selected groups
             nodes = [n for n in nodes if n.get("isSelected", True)]
@@ -373,18 +423,20 @@ class AgGridReturn(Mapping):
 
         # Check if response has groups
         has_groups = any(n.get("group", False) for n in nodes)
-        
+
         if has_groups:
             # Additional safety check: ensure we have leaf nodes with parent paths
             leaf_nodes = [n for n in nodes if not n.get("group", False)]
             leaf_with_parent_path = [n for n in leaf_nodes if n.get("parentPath")]
-            
+
             if leaf_with_parent_path:
                 return self._process_grouped_response(nodes)
             else:
                 # Has groups but no proper parent paths - fall back to regular data
-                print("Warning: Grouped data detected but no parentPath found in leaf nodes. Falling back to regular data.")
-        
+                print(
+                    "Warning: Grouped data detected but no parentPath found in leaf nodes. Falling back to regular data."
+                )
+
         # No groups or invalid grouped data - return single group with all data
         fallback_data = self._get_data(only_selected)
         return [{(): fallback_data}]
@@ -392,7 +444,7 @@ class AgGridReturn(Mapping):
     @property
     def dataGroups(self):
         """
-        Returns grouped rows as a dictionary where keys are tuples of 
+        Returns grouped rows as a dictionary where keys are tuples of
         groupby strings and values are pandas.DataFrame.
         """
         return self._get_data_groups(only_selected=False)
@@ -400,7 +452,7 @@ class AgGridReturn(Mapping):
     @property
     def selected_dataGroups(self):
         """
-        Returns selected rows as a dictionary where keys are tuples of 
+        Returns selected rows as a dictionary where keys are tuples of
         grouped column names and values are pandas.DataFrame.
         """
         return self._get_data_groups(only_selected=True)
@@ -411,13 +463,15 @@ class AgGridReturn(Mapping):
         Returns selected rows as a DataFrame.
         If there are grouped rows, returns a dict of {key: pd.DataFrame}.
         """
-        nodes = self.grid_response.get('nodes', [])
-        selected_items = pd.DataFrame([n.get('data', None) for n in nodes if n.get('isSelected', None) is True])
+        nodes = self.grid_response.get("nodes", [])
+        selected_items = pd.DataFrame(
+            [n.get("data", None) for n in nodes if n.get("isSelected", None) is True]
+        )
 
         if selected_items.empty:
             return None
 
-        # Set pandas index if available
+        # Set pandas index if available and remove the internal column
         if "::auto_unique_id::" in selected_items.columns:
             selected_items.set_index("::auto_unique_id::", inplace=True)
             selected_items.index.name = "index"
@@ -439,34 +493,40 @@ class AgGridReturn(Mapping):
             return getattr(self, key)
         except AttributeError:
             pass
-        
+
         # Try to get from grid_response
-        grid_response = self.__dict__.get('grid_response', {})
+        grid_response = self.__dict__.get("grid_response", {})
         if isinstance(grid_response, dict) and key in grid_response:
             return grid_response[key]
-        
+
         # Fall back to __dict__ access
         return self.__dict__[key]
 
     def __iter__(self):
         """Iterate over public attributes."""
-        return (name for name, _ in inspect.getmembers(self) if not name.startswith("_"))
+        return (
+            name for name, _ in inspect.getmembers(self) if not name.startswith("_")
+        )
 
     def __len__(self):
         """Return number of public attributes."""
-        return len([name for name, _ in inspect.getmembers(self) if not name.startswith("_")])
+        return len(
+            [name for name, _ in inspect.getmembers(self) if not name.startswith("_")]
+        )
 
     def keys(self):
         """Return all available keys (attributes + grid_response keys)."""
         # Get public attribute names
-        attr_keys = [name for name, _ in inspect.getmembers(self) if not name.startswith("_")]
-        
+        attr_keys = [
+            name for name, _ in inspect.getmembers(self) if not name.startswith("_")
+        ]
+
         # Get grid_response keys for backward compatibility
-        grid_response = self.__dict__.get('grid_response', {})
+        grid_response = self.__dict__.get("grid_response", {})
         if isinstance(grid_response, dict):
             grid_keys = [k for k in grid_response.keys() if k not in attr_keys]
             return attr_keys + grid_keys
-        
+
         return attr_keys
 
     def values(self):
